import copy
import json
import logging
import platform
import sys
from http.client import HTTPResponse
from typing import Union
from urllib.error import HTTPError
from urllib.parse import urlencode, quote
from urllib.request import Request, urlopen

from . import version
from .errors import SCIMApiError, SCIMError
from .group import Group
from .groups import Groups
from .request import SCIMRequest
from .response import SCIMResponse
from .service_provider_configs import ServiceProviderConfigs
from .user import User
from .users import Users


class SCIMClient:
    _logger = logging.getLogger(__name__)

    production_base_url = "https://api.slack.com/scim/v1"
    schema_values = [
        "urn:scim:schemas:core:1.0",
        "urn:scim:schemas:extension:enterprise:1.0"
    ]

    def __init__(
        self,
        token: str,
        base_url: str = production_base_url
    ):
        """Slack SCIM API Client

        https://api.slack.com/scim

        As this class doesn't have any other state apart from the token and base_Url, if your app is fine with
        reusing the same settings for requests, you can use an instance as singleton in your app.s

        :param token: An OAuth token with the admin scope is required to access the SCIM API.
            The simplest way to acquire this token is for an admin on a Workspace to create a new Slack app,
            add the admin OAuth scope, install the app, and use the generated token.
        :param base_url: the default one is the production URL. If you want to use proxy URL or test server URL,
            Give the URL to this parameter.
        """
        self.token: str = token
        self.base_url: str = base_url

    def __repr__(self):
        d: dict = {"token": "(redacted)", "base_url": self.base_url}
        return f"<slack_scim.{self.__class__.__name__}: {d}>"

    # ----------------------------------------------
    # User Management
    # ----------------------------------------------

    def create_user(
        self,
        user: Union[dict, User]
    ) -> User:
        """Creates a new user.

        https://api.slack.com/scim#users

        :param user: if you give a dict value here, be noted that keys must be camel-cased,
            not your familiar snake-case style.
        :return: API response
        :raise SCIMApiError: When getting an error code with unsuccessful HTTP status from Slack
        """
        req = SCIMRequest(
            token=self.token,
            http_method="POST",
            url=f"{self.base_url}/Users",
            json_body=self._to_user_dict(user),
        )
        resp = self.api_call(req)
        if resp.is_success():
            return User.from_dict(json.loads(resp.body))
        else:
            raise SCIMApiError.from_response(resp)

    def patch_user(
        self,
        id: str,
        user: Union[dict, User]
    ) -> User:
        """Partially updates a user.

        https://api.slack.com/scim#users

        :param id: user ID
        :param user: if you give a dict value here, be noted that keys must be camel-cased,
            not your familiar snake-case style.
        :return: API response
        :raise SCIMApiError: When getting an error code with unsuccessful HTTP status from Slack
        """
        id = self._ensure_user_id(id, user)
        req = SCIMRequest(
            token=self.token,
            http_method="PATCH",
            url=f"{self.base_url}/Users/{quote(id)}",
            json_body=self._to_user_dict(user),
        )
        resp = self.api_call(req)
        if resp.is_success():
            return User.from_dict(json.loads(resp.body)) if resp.body else None
        else:
            raise SCIMApiError.from_response(resp)

    def update_user(
        self,
        id: str,
        user: Union[dict, User],
    ) -> User:
        """Overwrites the whole user data.

        https://api.slack.com/scim#users

        :param id: user ID
        :param user: if you give a dict value here, be noted that keys must be camel-cased,
            not your familiar snake-case style.
        :return: API response
        :raise SCIMApiError: When getting an error code with unsuccessful HTTP status from Slack
        """
        id = self._ensure_user_id(id, user)
        req = SCIMRequest(
            token=self.token,
            http_method="PUT",
            url=f"{self.base_url}/Users/{quote(id)}",
            json_body=self._to_user_dict(user),
        )
        resp = self.api_call(req)
        if resp.status == 200:
            return User.from_dict(json.loads(resp.body)) if resp.body else None
        else:
            raise SCIMApiError.from_response(resp)

    def delete_user(
        self,
        id: str,
    ) -> User:
        """Deletes a user.

        https://api.slack.com/scim#users

        :param id: user ID
        :return: API response
        :raise SCIMApiError: When getting an error code with unsuccessful HTTP status from Slack
        """
        req = SCIMRequest(
            token=self.token,
            http_method="DELETE",
            url=f"{self.base_url}/Users/{quote(id)}",
        )
        resp = self.api_call(req)
        if not resp.is_success():
            raise SCIMApiError.from_response(resp)

    def read_user(
        self,
        id: str,
    ) -> User:
        """Finds a user by user ID.

        https://api.slack.com/scim#users

        :param id: user ID
        :return: API response
        :raise SCIMApiError: When getting an error code with unsuccessful HTTP status from Slack
        """
        req = SCIMRequest(
            token=self.token,
            http_method="GET",
            url=f"{self.base_url}/Users/{quote(id)}",
        )
        resp = self.api_call(req)
        if resp.is_success():
            return User.from_dict(json.loads(resp.body)) if resp.body else None
        else:
            raise SCIMApiError.from_response(resp)

    def search_users(
        self,
        *,
        filter: str = None,
        count: int = None,
        start_index: int = None
    ) -> Users:
        """Searches the users matching the given filter.

        https://api.slack.com/scim#users

        :param filter: https://api.slack.com/scim#filter
        :param count: the number of results to return in a response
        :param start_index: the index to fetch as the first item
        :return: API response
        :raise SCIMApiError: When getting an error code with unsuccessful HTTP status from Slack
        """
        query = {}
        if filter:
            query["filter"] = filter
        if count:
            query["count"] = count
        if start_index:
            query["startIndex"] = start_index
        req = SCIMRequest(
            token=self.token,
            http_method="GET",
            url=f"{self.base_url}/Users",
            query_params=query
        )
        resp = self.api_call(req)
        if resp.is_success():
            return Users.from_dict(json.loads(resp.body)) if resp.body else None
        else:
            raise SCIMApiError.from_response(resp)

    def _ensure_user_id(self, id: str, user: User) -> str:
        if id:
            return id
        else:
            if user:
                return user.id
            else:
                raise SCIMError("User ID is missing for update_user call")

    def _to_user_dict(self, user: Union[dict, User]) -> dict:
        if user:
            user_dict: dict = user if isinstance(user, dict) else user.to_dict()
            user_dict["schemas"] = self.schema_values
            return user_dict
        else:
            return None

    # ----------------------------------------------
    # Group Management
    # ----------------------------------------------

    def create_group(
        self,
        group: Union[dict, Group]
    ) -> Group:
        """Creates a new group.

        https://api.slack.com/scim#groups

        :param user: if you give a dict value here, be noted that keys must be camel-cased,
            not your familiar snake-case style.
        :return: API response
        :raise SCIMApiError: When getting an error code with unsuccessful HTTP status from Slack
        """

        req = SCIMRequest(
            token=self.token,
            http_method="POST",
            url=f"{self.base_url}/Groups",
            json_body=self._to_group_dict(group),
        )
        resp = self.api_call(req)
        if resp.is_success():
            return Group.from_dict(json.loads(resp.body)) if resp.body else None
        else:
            raise SCIMApiError.from_response(resp)

    def patch_group(
        self,
        id: str,
        group: Union[dict, Group]
    ) -> Group:
        """Partially updates a group.

        https://api.slack.com/scim#groups

        :param id: group ID
        :param user: if you give a dict value here, be noted that keys must be camel-cased,
            not your familiar snake-case style.
        :return: API response
        :raise SCIMApiError: When getting an error code with unsuccessful HTTP status from Slack
        """
        id = self._ensure_group_id(id, group)
        req = SCIMRequest(
            token=self.token,
            http_method="PATCH",
            url=f"{self.base_url}/Groups/{quote(id)}",
            json_body=self._to_group_dict(group),
        )
        resp = self.api_call(req)
        if resp.is_success():
            return Group.from_dict(json.loads(resp.body)) if resp.body else None
        else:
            raise SCIMApiError.from_response(resp)

    def update_group(
        self,
        id: str,
        group: Union[dict, Group],
    ) -> Group:
        """Overwrites the whole group.

        https://api.slack.com/scim#groups

        :param id: group ID
        :param user: if you give a dict value here, be noted that keys must be camel-cased,
            not your familiar snake-case style.
        :return: API response
        :raise SCIMApiError: When getting an error code with unsuccessful HTTP status from Slack
        """
        id = self._ensure_group_id(id, group)
        req = SCIMRequest(
            token=self.token,
            http_method="PUT",
            url=f"{self.base_url}/Groups/{quote(id)}",
            json_body=self._to_group_dict(group),
        )
        resp = self.api_call(req)
        if resp.is_success():
            return Group.from_dict(json.loads(resp.body)) if resp.body else None
        else:
            raise SCIMApiError.from_response(resp)

    def delete_group(
        self,
        id: str,
    ) -> Group:
        """Deletes a group.

        https://api.slack.com/scim#groups

        :param id: group ID
        :return: API response
        :raise SCIMApiError: When getting an error code with unsuccessful HTTP status from Slack
        """
        req = SCIMRequest(
            token=self.token,
            http_method="DELETE",
            url=f"{self.base_url}/Groups/{quote(id)}",
        )
        resp = self.api_call(req)
        if not resp.is_success():
            raise SCIMApiError.from_response(resp)

    def read_group(
        self,
        id: str,
    ) -> Group:
        """Finds a group by group ID.

        https://api.slack.com/scim#groups

        :param id: group ID
        :return: API response
        :raise SCIMApiError: When getting an error code with unsuccessful HTTP status from Slack
        """
        req = SCIMRequest(
            token=self.token,
            http_method="GET",
            url=f"{self.base_url}/Groups/{quote(id)}",
        )
        resp = self.api_call(req)
        if resp.is_success():
            return Group.from_dict(json.loads(resp.body)) if resp.body else None
        else:
            raise SCIMApiError.from_response(resp)

    def search_groups(
        self,
        *,
        filter: str = None,
        count: int = None,
        start_index: int = None
    ) -> Groups:
        """Searches the groups matching the given filter.

        https://api.slack.com/scim#users

        :param filter: https://api.slack.com/scim#filter
        :param count: the number of results to return in a response
        :param start_index: the index to fetch as the first item
        :return: API response
        :raise SCIMApiError: When getting an error code with unsuccessful HTTP status from Slack
        """
        query = {}
        if filter:
            query["filter"] = filter
        if count:
            query["count"] = count
        if start_index:
            query["startIndex"] = start_index
        req = SCIMRequest(
            token=self.token,
            http_method="GET",
            url=f"{self.base_url}/Groups",
            query_params=query
        )
        resp = self.api_call(req)
        if resp.is_success():
            return Groups.from_dict(json.loads(resp.body)) if resp.body else None
        else:
            raise SCIMApiError.from_response(resp)

    def _ensure_group_id(self, id: str, group: Group) -> str:
        if id:
            return id
        else:
            if group:
                return group.id
            else:
                raise SCIMError("Group ID is missing for update_group call")

    def _to_group_dict(self, group: Union[dict, Group]) -> dict:
        if group:
            group_dict: dict = group if isinstance(group, dict) else group.to_dict()
            group_dict["schemas"] = self.schema_values
            return group_dict
        else:
            return None

    # ----------------------------------------------
    # ServiceProviderConfigs
    # ----------------------------------------------

    def get_service_provider_configs(self) -> ServiceProviderConfigs:
        """Fetches ServiceProviderConfigs

        :raise SCIMApiError: When getting an error code with unsuccessful HTTP status from Slack
        :return: API response
        """
        req = SCIMRequest(
            token=self.token,
            http_method="GET",
            url=f"{self.base_url}/ServiceProviderConfigs",
        )
        resp = self.api_call(req)
        if resp.is_success():
            return ServiceProviderConfigs.from_dict(json.loads(resp.body)) if resp.body else None
        else:
            raise SCIMApiError.from_response(resp)

    # ----------------------------------------------
    # HTTP Client
    # ----------------------------------------------

    def api_call(self, api_request: SCIMRequest) -> SCIMResponse:
        """A general method to call the Slack SCIM APIs

        :param api_request: API request information
        :return: API response
        :raise Exception: only when unexpected errors occur,
            never raises exceptions when getting an error code
            with unsuccessful HTTP status from Slack
        """
        http_method = api_request.http_method.upper()
        url = api_request.url
        if api_request.query_params:
            params = copy.copy(api_request.query_params)
            if http_method == "GET" and api_request.body_params:
                params.update(api_request.body_params)
            q = urlencode(params)
            url = url + (f"&{q}" if "?" in url else f"?{q}")
            pass

        headers = copy.copy(api_request.headers)
        headers["Authorization"] = f"Bearer {api_request.token}"
        headers["User-Agent"] = self._build_user_agent()

        # The SCIM API never handles binary data
        if api_request.json_body:
            body: dict = self._to_non_null_dict(copy.copy(api_request.json_body))
            req_body: str = None if http_method == "GET" else json.dumps(body)
            headers["Content-Type"] = "application/json;charset=utf-8"
        else:
            req_body: str = None if http_method == "GET" else urlencode(api_request.body_params)
            headers["Content-Type"] = "application/x-www-form-urlencoded;charset=utf-8"

        req_data: bytes = req_body.encode("utf-8") if req_body else None

        try:
            http_request: Request = Request(
                method=http_method,
                url=url,
                data=req_data,
                headers=headers
            )
            self._debug_log_request(http_request, req_body)

            http_response: HTTPResponse = urlopen(http_request)
            charset: str = http_response.headers.get_content_charset()
            raw_body: bytes = http_response.read()
            resp_body: str = raw_body.decode(charset) if raw_body else None

            api_response = SCIMResponse(
                status=http_response.status,
                reason=http_response.reason,
                headers=http_response.headers,
                body=resp_body,
            )
            self._debug_log_completion(http_request, api_response)
            return api_response

        except HTTPError as e:
            # TODO: rate limit
            api_response = SCIMResponse(
                status=e.code,
                reason=e.reason,
                headers=e.headers,
                body=e.read().decode("utf-8"),
            )
            self._debug_log_completion(http_request, api_response)
            return api_response
        except Exception as e:
            self._logger.error(f"Failed to send a request to Slack SCIM API server: {e}")
            raise e

    def _to_non_null_dict(self, d: Union[dict, any]) -> dict:
        """Recursively converts an object to dict"""
        result = {}
        if isinstance(d, dict):
            for key, value in d.items():
                if value is None:
                    continue
                if isinstance(value, dict):
                    value = self._to_non_null_dict(value)
                    result[key] = value
                elif isinstance(value, list):
                    new_list = []
                    for v in value:
                        v = self._to_non_null_dict(v)
                        if v:
                            new_list.append(v)
                    result[key] = new_list
                else:
                    result[key] = value
            return result
        elif getattr(d, "to_dict", None):
            return d.to_dict()
        else:
            return d

    def _debug_log_request(self, req: Request, body: str):
        if self._logger.level <= logging.DEBUG:
            headers_part = "\n".join([
                f"{k.lower()}: (redacted)" if k.lower() == "authorization" else f"{k.lower()}: {v}"
                for k, v in req.headers.items()
            ])
            message = f"*** SCIM API Request ***\n" \
                      f"{req.method} {req.full_url}\n" \
                      f"{headers_part}\n\n" \
                      f"{body or ''}\n"
            self._logger.debug(message)

    def _debug_log_completion(self, req: Request, resp: SCIMResponse):
        if self._logger.level <= logging.DEBUG:
            headers_part = "\n".join([f"{k}: {v}" for k, v in resp.headers.items()])
            message = f"*** SCIM API Response ***\n" \
                      f"{req.method} {req.full_url}\n" \
                      f"{resp.status} {resp.reason}\n" \
                      f"{headers_part}\n\n" \
                      f"{resp.body or ''}\n"
            self._logger.debug(message)

    def _build_user_agent(self):
        client = "{0}/{1}".format("slack_scim", version.__version__)
        python_version = "Python/{v.major}.{v.minor}.{v.micro}".format(v=sys.version_info)
        system_info = "{0}/{1}".format(platform.system(), platform.release())
        user_agent_string = " ".join([python_version, client, system_info])
        return user_agent_string



# The default is v1
from .v1.client import SCIMClient
from .v1.errors import SCIMApiError
from .v1.group import Group, Member as GroupMember
from .v1.groups import Groups
from .v1.service_provider_configs import ServiceProviderConfigs
from .v1.user import User
from .v1.users import Users
from .v1.request import SCIMRequest
from .v1.response import SCIMResponse


import pkg_resources  # part of setuptools


def load_package_version():
    package = pkg_resources.require("slack_scim")
    if package and len(package) > 0:
        return package[0].version or "0.0.0"


__version__ = load_package_version()



# To use this code, make sure you
#
#     import json
#
# and then, to convert JSON from a string, do
#
#     result = service_provider_configs_from_dict(json.loads(json_string))

from typing import Optional, Any, List, TypeVar, Callable, Type, cast


T = TypeVar("T")


def from_str(x: Any) -> str:
    assert isinstance(x, str)
    return x


def from_none(x: Any) -> Any:
    assert x is None
    return x


def from_union(fs, x):
    for f in fs:
        try:
            return f(x)
        except:
            pass
    assert False


def from_bool(x: Any) -> bool:
    assert isinstance(x, bool)
    return x


def from_int(x: Any) -> int:
    assert isinstance(x, int) and not isinstance(x, bool)
    return x


def from_list(f: Callable[[Any], T], x: Any) -> List[T]:
    assert isinstance(x, list)
    return [f(y) for y in x]


def to_class(c: Type[T], x: Any) -> dict:
    assert isinstance(x, c)
    return cast(Any, x).to_dict()


class AuthenticationScheme:
    description: Optional[str]
    name: Optional[str]
    primary: Optional[bool]
    spec_url: Optional[str]
    type: Optional[str]

    def __init__(self, description: Optional[str], name: Optional[str], primary: Optional[bool], spec_url: Optional[str], type: Optional[str]) -> None:
        self.description = description
        self.name = name
        self.primary = primary
        self.spec_url = spec_url
        self.type = type

    @staticmethod
    def from_dict(obj: Any) -> 'AuthenticationScheme':
        assert isinstance(obj, dict)
        description = from_union([from_str, from_none], obj.get("description"))
        name = from_union([from_str, from_none], obj.get("name"))
        primary = from_union([from_bool, from_none], obj.get("primary"))
        spec_url = from_union([from_str, from_none], obj.get("specUrl"))
        type = from_union([from_str, from_none], obj.get("type"))
        return AuthenticationScheme(description, name, primary, spec_url, type)

    def to_dict(self) -> dict:
        result: dict = {}
        result["description"] = from_union([from_str, from_none], self.description)
        result["name"] = from_union([from_str, from_none], self.name)
        result["primary"] = from_union([from_bool, from_none], self.primary)
        result["specUrl"] = from_union([from_str, from_none], self.spec_url)
        result["type"] = from_union([from_str, from_none], self.type)
        return result


class Bulk:
    max_operations: Optional[int]
    max_payload_size: Optional[int]
    supported: Optional[bool]

    def __init__(self, max_operations: Optional[int], max_payload_size: Optional[int], supported: Optional[bool]) -> None:
        self.max_operations = max_operations
        self.max_payload_size = max_payload_size
        self.supported = supported

    @staticmethod
    def from_dict(obj: Any) -> 'Bulk':
        assert isinstance(obj, dict)
        max_operations = from_union([from_int, from_none], obj.get("maxOperations"))
        max_payload_size = from_union([from_int, from_none], obj.get("maxPayloadSize"))
        supported = from_union([from_bool, from_none], obj.get("supported"))
        return Bulk(max_operations, max_payload_size, supported)

    def to_dict(self) -> dict:
        result: dict = {}
        result["maxOperations"] = from_union([from_int, from_none], self.max_operations)
        result["maxPayloadSize"] = from_union([from_int, from_none], self.max_payload_size)
        result["supported"] = from_union([from_bool, from_none], self.supported)
        return result


class ChangePassword:
    supported: Optional[bool]

    def __init__(self, supported: Optional[bool]) -> None:
        self.supported = supported

    @staticmethod
    def from_dict(obj: Any) -> 'ChangePassword':
        assert isinstance(obj, dict)
        supported = from_union([from_bool, from_none], obj.get("supported"))
        return ChangePassword(supported)

    def to_dict(self) -> dict:
        result: dict = {}
        result["supported"] = from_union([from_bool, from_none], self.supported)
        return result


class Filter:
    max_results: Optional[int]
    supported: Optional[bool]

    def __init__(self, max_results: Optional[int], supported: Optional[bool]) -> None:
        self.max_results = max_results
        self.supported = supported

    @staticmethod
    def from_dict(obj: Any) -> 'Filter':
        assert isinstance(obj, dict)
        max_results = from_union([from_int, from_none], obj.get("maxResults"))
        supported = from_union([from_bool, from_none], obj.get("supported"))
        return Filter(max_results, supported)

    def to_dict(self) -> dict:
        result: dict = {}
        result["maxResults"] = from_union([from_int, from_none], self.max_results)
        result["supported"] = from_union([from_bool, from_none], self.supported)
        return result


class ServiceProviderConfigs:
    authentication_schemes: Optional[List[AuthenticationScheme]]
    bulk: Optional[Bulk]
    change_password: Optional[ChangePassword]
    etag: Optional[ChangePassword]
    filter: Optional[Filter]
    patch: Optional[ChangePassword]
    sort: Optional[ChangePassword]
    xml_data_format: Optional[ChangePassword]

    def __init__(self, authentication_schemes: Optional[List[AuthenticationScheme]], bulk: Optional[Bulk], change_password: Optional[ChangePassword], etag: Optional[ChangePassword], filter: Optional[Filter], patch: Optional[ChangePassword], sort: Optional[ChangePassword], xml_data_format: Optional[ChangePassword]) -> None:
        self.authentication_schemes = authentication_schemes
        self.bulk = bulk
        self.change_password = change_password
        self.etag = etag
        self.filter = filter
        self.patch = patch
        self.sort = sort
        self.xml_data_format = xml_data_format

    @staticmethod
    def from_dict(obj: Any) -> 'ServiceProviderConfigs':
        assert isinstance(obj, dict)
        authentication_schemes = from_union([lambda x: from_list(AuthenticationScheme.from_dict, x), from_none], obj.get("authenticationSchemes"))
        bulk = from_union([Bulk.from_dict, from_none], obj.get("bulk"))
        change_password = from_union([ChangePassword.from_dict, from_none], obj.get("changePassword"))
        etag = from_union([ChangePassword.from_dict, from_none], obj.get("etag"))
        filter = from_union([Filter.from_dict, from_none], obj.get("filter"))
        patch = from_union([ChangePassword.from_dict, from_none], obj.get("patch"))
        sort = from_union([ChangePassword.from_dict, from_none], obj.get("sort"))
        xml_data_format = from_union([ChangePassword.from_dict, from_none], obj.get("xmlDataFormat"))
        return ServiceProviderConfigs(authentication_schemes, bulk, change_password, etag, filter, patch, sort, xml_data_format)

    def to_dict(self) -> dict:
        result: dict = {}
        result["authenticationSchemes"] = from_union([lambda x: from_list(lambda x: to_class(AuthenticationScheme, x), x), from_none], self.authentication_schemes)
        result["bulk"] = from_union([lambda x: to_class(Bulk, x), from_none], self.bulk)
        result["changePassword"] = from_union([lambda x: to_class(ChangePassword, x), from_none], self.change_password)
        result["etag"] = from_union([lambda x: to_class(ChangePassword, x), from_none], self.etag)
        result["filter"] = from_union([lambda x: to_class(Filter, x), from_none], self.filter)
        result["patch"] = from_union([lambda x: to_class(ChangePassword, x), from_none], self.patch)
        result["sort"] = from_union([lambda x: to_class(ChangePassword, x), from_none], self.sort)
        result["xmlDataFormat"] = from_union([lambda x: to_class(ChangePassword, x), from_none], self.xml_data_format)
        return result


def service_provider_configs_from_dict(s: Any) -> ServiceProviderConfigs:
    return ServiceProviderConfigs.from_dict(s)


def service_provider_configs_to_dict(x: ServiceProviderConfigs) -> Any:
    return to_class(ServiceProviderConfigs, x)



import json
from typing import Dict

from .response import SCIMResponse


class SCIMError(Exception):
    def __init__(self, errors: Dict[str, any]):
        """Represents a general error in this library

        :param errors: Specific error information
        """
        super()
        self.errors = errors

    def to_dict(self) -> dict:
        result: dict = {}
        result["errors"] = self.errors
        return result

    def __repr__(self):
        return f"<slack_scim.{self.__class__.__name__}: {self.to_dict()}>"


class SCIMApiError(SCIMError):
    def __init__(
        self,
        *,
        status: int,
        headers: Dict[str, str],
        errors: Dict[str, any],
    ):
        """Exception representing an error returned by Slack

        :param status: HTTP status code
        :param headers: All the response headers
        :param errors: Error messages returned by Slack
        """
        super(SCIMApiError, self).__init__(errors)
        self.status = status
        self.headers = headers
        self.errors = errors

    @classmethod
    def from_response(cls, resp: SCIMResponse) -> "SCIMApiError":
        errors = {}
        if resp.body:
            b = json.loads(resp.body)
            errors = b["Errors"] if "Errors" in b else {}
        return SCIMApiError(
            status=resp.status,
            headers=resp.headers,
            errors=errors,
        )

    def to_dict(self) -> dict:
        result: dict = {}
        result["status"] = self.status
        result["headers"] = self.headers
        result["errors"] = self.errors
        return result

    def __str__(self):
        return str(self.to_dict())

    def __repr__(self):
        return f"<slack_scim.{self.__class__.__name__}: {self.to_dict()}>"



# To use this code, make sure you
#
#     import json
#
# and then, to convert JSON from a string, do
#
#     result = group_from_dict(json.loads(json_string))

from typing import Optional, Any, List, TypeVar, Callable, Type, cast


T = TypeVar("T")


def from_str(x: Any) -> str:
    assert isinstance(x, str)
    return x


def from_none(x: Any) -> Any:
    assert x is None
    return x


def from_union(fs, x):
    for f in fs:
        try:
            return f(x)
        except:
            pass
    assert False


def from_list(f: Callable[[Any], T], x: Any) -> List[T]:
    assert isinstance(x, list)
    return [f(y) for y in x]


def to_class(c: Type[T], x: Any) -> dict:
    assert isinstance(x, c)
    return cast(Any, x).to_dict()


class Member:
    display: Optional[str]
    value: Optional[str]

    def __init__(self, display: Optional[str], value: Optional[str]) -> None:
        self.display = display
        self.value = value

    @staticmethod
    def from_dict(obj: Any) -> 'Member':
        assert isinstance(obj, dict)
        display = from_union([from_str, from_none], obj.get("display"))
        value = from_union([from_str, from_none], obj.get("value"))
        return Member(display, value)

    def to_dict(self) -> dict:
        result: dict = {}
        result["display"] = from_union([from_str, from_none], self.display)
        result["value"] = from_union([from_str, from_none], self.value)
        return result


class Meta:
    created: Optional[str]
    location: Optional[str]

    def __init__(self, created: Optional[str], location: Optional[str]) -> None:
        self.created = created
        self.location = location

    @staticmethod
    def from_dict(obj: Any) -> 'Meta':
        assert isinstance(obj, dict)
        created = from_union([from_str, from_none], obj.get("created"))
        location = from_union([from_str, from_none], obj.get("location"))
        return Meta(created, location)

    def to_dict(self) -> dict:
        result: dict = {}
        result["created"] = from_union([from_str, from_none], self.created)
        result["location"] = from_union([from_str, from_none], self.location)
        return result


class Group:
    display_name: Optional[str]
    id: Optional[str]
    members: Optional[List[Member]]
    meta: Optional[Meta]
    schemas: Optional[List[str]]

    def __init__(self, display_name: Optional[str], id: Optional[str], members: Optional[List[Member]], meta: Optional[Meta], schemas: Optional[List[str]]) -> None:
        self.display_name = display_name
        self.id = id
        self.members = members
        self.meta = meta
        self.schemas = schemas

    @staticmethod
    def from_dict(obj: Any) -> 'Group':
        assert isinstance(obj, dict)
        display_name = from_union([from_str, from_none], obj.get("displayName"))
        id = from_union([from_str, from_none], obj.get("id"))
        members = from_union([lambda x: from_list(Member.from_dict, x), from_none], obj.get("members"))
        meta = from_union([Meta.from_dict, from_none], obj.get("meta"))
        schemas = from_union([lambda x: from_list(from_str, x), from_none], obj.get("schemas"))
        return Group(display_name, id, members, meta, schemas)

    def to_dict(self) -> dict:
        result: dict = {}
        result["displayName"] = from_union([from_str, from_none], self.display_name)
        result["id"] = from_union([from_str, from_none], self.id)
        result["members"] = from_union([lambda x: from_list(lambda x: to_class(Member, x), x), from_none], self.members)
        result["meta"] = from_union([lambda x: to_class(Meta, x), from_none], self.meta)
        result["schemas"] = from_union([lambda x: from_list(from_str, x), from_none], self.schemas)
        return result


def group_from_dict(s: Any) -> Group:
    return Group.from_dict(s)


def group_to_dict(x: Group) -> Any:
    return to_class(Group, x)



from typing import Dict


class SCIMRequest():
    def __init__(
        self,
        *,
        token: str,
        http_method: str,
        url: str,
        query_params: Dict[str, str] = dict(),
        body_params: Dict[str, str] = dict(),
        json_body: Dict = dict(),
        headers: Dict[str, str] = dict(),
    ):
        self.token = token
        self.http_method = http_method
        self.url = url
        self.headers = headers
        self.query_params = query_params
        self.body_params = body_params
        self.json_body = json_body

    def to_dict(self) -> dict:
        return {
            "token": self.token,
            "http_method": self.http_method,
            "url": self.url,
            "headers": self.headers,
            "query_params": self.query_params,
            "body_params": self.body_params,
            "json_body": self.json_body,
        }

    def to_printable_dict(self) -> dict:
        result: dict = self.to_dict()
        result["token"] = "(redacted)"
        return result

    def __str__(self):
        return str(self.to_printable_dict())

    def __repr__(self):
        return f"<slack_scim.{self.__class__.__name__}: {self.to_printable_dict()}>"



#!/usr/bin/env python
from glob import glob
from os.path import splitext, basename

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="slack-scim",
    version="1.1.0",
    author="Kazuhiro Sera",
    author_email="seratch@gmail.com",
    description="Slack SCIM API Client",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/seratch/python-slack-scim",
    packages=setuptools.find_packages("src"),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    setup_requires=["pytest-runner==5.2"],
    tests_require=["pytest>=5,<6"],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
